@@ -6,9 +6,10 @@ from functools import reduce, singledispatch
 
 import opt_einsum
 
+from funsor.domains import Bint
 from funsor.interpreter import gensym
-from funsor.tensor import Einsum, Tensor, get_default_prototype
-from funsor.terms import Binary, Funsor, Lambda, Reduce, Unary, Variable, Bint
+from funsor.tensor import Tensor, get_default_prototype
+from funsor.terms import Binary, Finitary, Funsor, Lambda, Reduce, Unary, Variable
 
 from . import ops
 
@@ -37,7 +38,7 @@ def affine_inputs(fn):
     :return: A set of input names wrt which ``fn`` is affine.
     :rtype: frozenset
     """
-    result = getattr(fn, '_affine_inputs', None)
+    result = getattr(fn, "_affine_inputs", None)
     if result is None:
         result = fn._affine_inputs = _affine_inputs(fn)
     return result
@@ -60,7 +61,9 @@ def _(fn):
 
 @affine_inputs.register(Unary)
 def _(fn):
-    if fn.op in (ops.neg, ops.add) or isinstance(fn.op, ops.ReshapeOp):
+    if fn.op in (ops.neg, ops.sum) or isinstance(
+        fn.op, (ops.ReshapeOp, ops.GetsliceOp)
+    ):
         return affine_inputs(fn.arg)
     return frozenset()
 
@@ -91,12 +94,12 @@ def _(fn):
     return affine_inputs(fn.arg) - fn.reduced_vars
 
 
-@affine_inputs.register(Einsum)
+@affine_inputs.register(Finitary[ops.EinsumOp, tuple])
 def _(fn):
     # This is simply a multiary version of the above Binary(ops.mul, ...) case.
     results = []
-    for i, x in enumerate(fn.operands):
-        others = fn.operands[:i] + fn.operands[i+1:]
+    for i, x in enumerate(fn.args):
+        others = fn.args[:i] + fn.args[i + 1 :]
         other_inputs = reduce(ops.or_, map(_real_inputs, others), frozenset())
         results.append(affine_inputs(x) - other_inputs)
     # This multilinear case introduces incompleteness, since some vars
@@ -114,7 +117,7 @@ def extract_affine(fn):
 
         x = ...
         const, coeffs = extract_affine(x)
-        y = sum(Einsum(eqn, (coeff, Variable(var, coeff.output)))
+        y = sum(Einsum(eqn, coeff, Variable(var, coeff.output))
                 for var, (coeff, eqn) in coeffs.items())
         assert_close(y, x)
         assert frozenset(coeffs) == affine_inputs(x)
@@ -125,7 +128,7 @@ def extract_affine(fn):
 
     The affine approximation is computed by ev evaluating ``fn`` at
     zero and each basis vector. To improve performance, users may want to run
-    under the :func:`~funsor.memoize.memoize` interpretation.
+    under the :func:`~funsor.interpretations.Memoize` interpretation.
 
     :param Funsor fn: A funsor that is affine wrt the (add,mul) semiring in
         some subset of its inputs.
@@ -143,7 +146,7 @@ def extract_affine(fn):
     const = fn(**zeros)
 
     # Determine linear coefficients by evaluating fn on basis vectors.
-    name = gensym('probe')
+    name = gensym("probe")
     coeffs = OrderedDict()
     for k, v in inputs.items():
         dim = v.num_elements
@@ -151,10 +154,10 @@ def extract_affine(fn):
         subs = zeros.copy()
         subs[k] = Tensor(ops.new_eye(prototype, (dim,)).reshape((dim,) + v.shape))[var]
         coeff = Lambda(var, fn(**subs) - const).reshape(v.shape + const.shape)
-        inputs1 = ''.join(map(opt_einsum.get_symbol, range(len(coeff.shape))))
-        inputs2 = inputs1[:len(v.shape)]
-        output = inputs1[len(v.shape):]
-        eqn = '{},{}->{}'.format(inputs1, inputs2, output)
+        inputs1 = "".join(map(opt_einsum.get_symbol, range(len(coeff.shape))))
+        inputs2 = inputs1[: len(v.shape)]
+        output = inputs1[len(v.shape) :]
+        eqn = "{},{}->{}".format(inputs1, inputs2, output)
         coeffs[k] = coeff, eqn
     return const, coeffs
 
