@@ -12,14 +12,14 @@ def einsum(equation, *operands):
     """
     if get_backend() != "jax":
         # NB: rename symbols to support NumPy, which allow only symbols a-z.
-        symbols = sorted(set(equation) - set(',->'))
-        rename = dict(zip(symbols, 'abcdefghijklmnopqrstuvwxyz'))
-        equation = ''.join(rename.get(s, s) for s in equation)
+        symbols = sorted(set(equation) - set(",->"))
+        rename = dict(zip(symbols, "abcdefghijklmnopqrstuvwxyz"))
+        equation = "".join(rename.get(s, s) for s in equation)
 
-    inputs, output = equation.split('->')
+    inputs, output = equation.split("->")
     if inputs == output:
         return operands[0][...]  # create a new object
-    inputs = inputs.split(',')
+    inputs = inputs.split(",")
 
     shifts = []
     exp_operands = []
@@ -33,7 +33,9 @@ def einsum(equation, *operands):
         exp_operands.append(ops.exp(operand - shift))
 
         # permute shift to match output
-        shift = shift.reshape([size for size, dim in zip(operand.shape, dims) if dim in output])
+        shift = shift.reshape(
+            [size for size, dim in zip(operand.shape, dims) if dim in output]
+        )
         if len(shift.shape) > 0:
             shift = shift.reshape((1,) * (len(output) - shift.ndim) + shift.shape)
             dims = [dim for dim in dims if dim in output]
@@ -41,7 +43,7 @@ def einsum(equation, *operands):
             shift = ops.permute(shift, [dims.index(dim) for dim in output])
         shifts.append(shift)
 
-    result = ops.log(ops.einsum(equation, *exp_operands))
+    result = ops.log(ops.einsum(exp_operands, equation))
     return sum(shifts + [result])
 
 
