@@ -17,20 +17,23 @@ from multipledispatch.variadic import Variadic
 import funsor.ops as ops
 from funsor.cnf import Contraction
 from funsor.delta import Delta
-from funsor.domains import Domain, Bint, Real
+from funsor.domains import Bint, Domain, Real
 from funsor.gaussian import Gaussian
-from funsor.terms import Funsor, Number
 from funsor.tensor import Tensor
+from funsor.terms import Funsor, Number
 from funsor.util import get_backend
 
 
 @contextlib.contextmanager
-def xfail_if_not_implemented(msg="Not implemented"):
+def xfail_if_not_implemented(msg="Not implemented", *, match=None):
     try:
         yield
     except NotImplementedError as e:
+        if match is not None and match not in str(e):
+            raise e from None
         import pytest
-        pytest.xfail(reason='{}:\n{}'.format(msg, e))
+
+        pytest.xfail(reason="{}:\n{}".format(msg, e))
 
 
 @contextlib.contextmanager
@@ -39,23 +42,41 @@ def xfail_if_not_found(msg="Not implemented"):
         yield
     except AttributeError as e:
         import pytest
-        pytest.xfail(reason='{}:\n{}'.format(msg, e))
+
+        pytest.xfail(reason="{}:\n{}".format(msg, e))
 
 
-class ActualExpected(namedtuple('LazyComparison', ['actual', 'expected'])):
+def requires_backend(*backends, reason=None):
+    import pytest
+
+    if reason is None:
+        reason = "Test requires backend {}".format(" or ".join(backends))
+    return pytest.mark.skipif(get_backend() not in backends, reason=reason)
+
+
+def excludes_backend(*backends, reason=None):
+    import pytest
+
+    if reason is None:
+        reason = "Test excludes backend {}".format(" and ".join(backends))
+    return pytest.mark.skipif(get_backend() in backends, reason=reason)
+
+
+class ActualExpected(namedtuple("LazyComparison", ["actual", "expected"])):
     """
     Lazy string formatter for test assertions.
     """
+
     def __repr__(self):
-        return '\n'.join(['Expected:', str(self.expected), 'Actual:', str(self.actual)])
+        return "\n".join(["Expected:", str(self.expected), "Actual:", str(self.actual)])
 
 
 def id_from_inputs(inputs):
     if isinstance(inputs, (dict, OrderedDict)):
         inputs = inputs.items()
     if not inputs:
-        return '()'
-    return ','.join(k + ''.join(map(str, d.shape)) for k, d in inputs)
+        return "()"
+    return ",".join(k + "".join(map(str, d.shape)) for k, d in inputs)
 
 
 @dispatch(object, object, Variadic[float])
@@ -76,7 +97,11 @@ def allclose(a, b, rtol=1e-05, atol=1e-08):
 
 
 def is_array(x):
-    return get_backend() != "torch" and ops.is_numeric_array(x)
+    if isinstance(x, Funsor):
+        return False
+    if get_backend() == "torch":
+        return False
+    return ops.is_numeric_array(x)
 
 
 def assert_close(actual, expected, atol=1e-6, rtol=1e-6):
@@ -84,35 +109,48 @@ def assert_close(actual, expected, atol=1e-6, rtol=1e-6):
     if is_array(actual):
         assert is_array(expected), msg
     elif isinstance(actual, Tensor) and is_array(actual.data):
-        assert isinstance(expected, Tensor) and is_array(expected.data)
-    elif isinstance(actual, Contraction) and isinstance(actual.terms[0], Tensor) \
-            and is_array(actual.terms[0].data):
-        assert isinstance(expected, Contraction) and is_array(expected.terms[0].data)
-    elif isinstance(actual, Gaussian) and is_array(actual.info_vec):
-        assert isinstance(expected, Gaussian) and is_array(expected.info_vec)
+        assert isinstance(expected, Tensor) and is_array(expected.data), msg
+    elif (
+        isinstance(actual, Contraction)
+        and isinstance(actual.terms[0], Tensor)
+        and is_array(actual.terms[0].data)
+    ):
+        assert isinstance(expected, Contraction) and is_array(
+            expected.terms[0].data
+        ), msg
+    elif isinstance(actual, Contraction) and isinstance(actual.terms[0], Delta):
+        assert isinstance(expected, Contraction) and isinstance(
+            expected.terms[0], Delta
+        ), msg
+    elif isinstance(actual, Gaussian):
+        assert isinstance(expected, Gaussian)
     else:
         assert type(actual) == type(expected), msg
 
     if isinstance(actual, Funsor):
-        assert isinstance(actual, Funsor)
-        assert isinstance(expected, Funsor)
+        assert isinstance(expected, Funsor), msg
         assert actual.inputs == expected.inputs, (actual.inputs, expected.inputs)
         assert actual.output == expected.output, (actual.output, expected.output)
 
     if isinstance(actual, (Number, Tensor)):
         assert_close(actual.data, expected.data, atol=atol, rtol=rtol)
     elif isinstance(actual, Delta):
-        assert frozenset(n for n, p in actual.terms) == frozenset(n for n, p in expected.terms)
+        assert frozenset(n for n, p in actual.terms) == frozenset(
+            n for n, p in expected.terms
+        )
         actual = actual.align(tuple(n for n, p in expected.terms))
-        for (actual_name, (actual_point, actual_log_density)), \
-                (expected_name, (expected_point, expected_log_density)) in \
-                zip(actual.terms, expected.terms):
+        for (actual_name, (actual_point, actual_log_density)), (
+            expected_name,
+            (expected_point, expected_log_density),
+        ) in zip(actual.terms, expected.terms):
             assert actual_name == expected_name
             assert_close(actual_point, expected_point, atol=atol, rtol=rtol)
             assert_close(actual_log_density, expected_log_density, atol=atol, rtol=rtol)
     elif isinstance(actual, Gaussian):
-        assert_close(actual.info_vec, expected.info_vec, atol=atol, rtol=rtol)
-        assert_close(actual.precision, expected.precision, atol=atol, rtol=rtol)
+        # Note white_vec and prec_sqrt are expected to agree only up to an
+        # orthogonal factor, but precision and info_vec should agree exactly.
+        assert_close(actual._info_vec, expected._info_vec, atol=atol, rtol=rtol)
+        assert_close(actual._precision, expected._precision, atol=atol, rtol=rtol)
     elif isinstance(actual, Contraction):
         assert actual.red_op == expected.red_op
         assert actual.bin_op == expected.bin_op
@@ -129,7 +167,7 @@ def assert_close(actual, expected, atol=1e-6, rtol=1e-6):
         if actual.dtype in (torch.long, torch.uint8, torch.bool):
             assert (actual == expected).all(), msg
         else:
-            eq = (actual == expected)
+            eq = actual == expected
             if eq.all():
                 return
             if eq.any():
@@ -141,20 +179,21 @@ def assert_close(actual, expected, atol=1e-6, rtol=1e-6):
             elif atol is not None:
                 assert diff.max() < atol, msg
     elif is_array(actual):
-        if isinstance(actual, (np.ndarray, np.generic)):
-            assert actual.dtype == expected.dtype, msg
-        else:
-            assert get_backend() == "jax"
+        if get_backend() == "jax":
             import jax
 
-            assert actual.dtype == jax.dtypes.canonicalize_dtype(expected.dtype), msg
+            assert jax.numpy.result_type(actual.dtype) == jax.numpy.result_type(
+                expected.dtype
+            ), msg
+        else:
+            assert actual.dtype == expected.dtype, msg
 
         assert actual.shape == expected.shape, msg
-        if actual.dtype in (np.int32, np.int64, np.uint8, np.bool):
+        if actual.dtype in (np.int32, np.int64, np.uint8, bool):
             assert (actual == expected).all(), msg
         else:
             actual, expected = np.asarray(actual), np.asarray(expected)
-            eq = (actual == expected)
+            eq = actual == expected
             if eq.all():
                 return
             if eq.any():
@@ -171,8 +210,18 @@ def assert_close(actual, expected, atol=1e-6, rtol=1e-6):
             assert diff < (atol + abs(expected)) * rtol, msg
         elif atol is not None:
             assert diff < atol, msg
+    elif isinstance(actual, dict):
+        assert isinstance(expected, dict)
+        assert set(actual) == set(expected)
+        for k, actual_v in actual.items():
+            assert_close(actual_v, expected[k], atol=atol, rtol=rtol)
+    elif isinstance(actual, tuple):
+        assert isinstance(expected, tuple)
+        assert len(actual) == len(expected)
+        for actual_v, expected_v in zip(actual, expected):
+            assert_close(actual_v, expected_v, atol=atol, rtol=rtol)
     else:
-        raise ValueError('cannot compare objects of type {}'.format(type(actual)))
+        raise ValueError("cannot compare objects of type {}".format(type(actual)))
 
 
 def check_funsor(x, inputs, output, data=None):
@@ -196,15 +245,16 @@ def check_funsor(x, inputs, output, data=None):
 
 def xfail_param(*args, **kwargs):
     import pytest
+
     return pytest.param(*args, marks=[pytest.mark.xfail(**kwargs)])
 
 
 def make_einsum_example(equation, fill=None, sizes=(2, 3)):
-    symbols = sorted(set(equation) - set(',->'))
+    symbols = sorted(set(equation) - set(",->"))
     sizes = {dim: size for dim, size in zip(symbols, itertools.cycle(sizes))}
-    inputs, outputs = equation.split('->')
-    inputs = inputs.split(',')
-    outputs = outputs.split(',')
+    inputs, outputs = equation.split("->")
+    inputs = inputs.split(",")
+    outputs = outputs.split(",")
     operands = []
     for dims in inputs:
         shape = tuple(sizes[dim] for dim in dims)
@@ -219,8 +269,9 @@ def make_einsum_example(equation, fill=None, sizes=(2, 3)):
         for inp, operand in zip(inputs, operands)
     ]
 
-    assert equation == \
-        ",".join(["".join(operand.inputs.keys()) for operand in funsor_operands]) + "->" + ",".join(outputs)
+    assert equation == ",".join(
+        ["".join(operand.inputs.keys()) for operand in funsor_operands]
+    ) + "->" + ",".join(outputs)
     return inputs, outputs, sizes, operands, funsor_operands
 
 
@@ -283,11 +334,16 @@ def random_scale_tril(*args):
         shape = args
 
     from funsor.distribution import BACKEND_TO_DISTRIBUTIONS_BACKEND
-    backend_dist = importlib.import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()]).dist
+
+    backend_dist = importlib.import_module(
+        BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()]
+    ).dist
 
     if get_backend() == "torch":
         data = randn(shape)
-        return backend_dist.transforms.transform_to(backend_dist.constraints.lower_cholesky)(data)
+        return backend_dist.transforms.transform_to(
+            backend_dist.constraints.lower_cholesky
+        )(data)
     else:
         data = randn(shape[:-2] + (shape[-1] * (shape[-1] + 1) // 2,))
         return backend_dist.biject_to(backend_dist.constraints.lower_cholesky)(data)
@@ -349,14 +405,16 @@ def random_tensor(inputs, output=Real):
     assert isinstance(inputs, OrderedDict)
     assert isinstance(output, Domain)
     shape = tuple(d.dtype for d in inputs.values()) + output.shape
-    if output.dtype == 'real':
+    if output.dtype == "real":
         data = randn(shape)
     else:
         num_elements = reduce(operator.mul, shape, 1)
         if backend == "torch":
             import torch
 
-            data = torch.multinomial(torch.ones(output.dtype), num_elements, replacement=True)
+            data = torch.multinomial(
+                torch.ones(output.dtype), num_elements, replacement=True
+            )
         else:
             data = np.random.choice(output.dtype, num_elements, replace=True)
         data = data.reshape(shape)
@@ -368,14 +426,15 @@ def random_gaussian(inputs):
     Creates a random :class:`funsor.gaussian.Gaussian` with given inputs.
     """
     assert isinstance(inputs, OrderedDict)
-    batch_shape = tuple(d.dtype for d in inputs.values() if d.dtype != 'real')
-    event_shape = (sum(d.num_elements for d in inputs.values() if d.dtype == 'real'),)
+    batch_shape = tuple(d.dtype for d in inputs.values() if d.dtype != "real")
+    event_shape = (sum(d.num_elements for d in inputs.values() if d.dtype == "real"),)
     prec_sqrt = randn(batch_shape + event_shape + event_shape)
     precision = ops.matmul(prec_sqrt, ops.transpose(prec_sqrt, -1, -2))
     precision = precision + 0.5 * ops.new_eye(precision, event_shape[:1])
+    prec_sqrt = ops.cholesky(precision)
     loc = randn(batch_shape + event_shape)
-    info_vec = ops.matmul(precision, ops.unsqueeze(loc, -1)).squeeze(-1)
-    return Gaussian(info_vec, precision, inputs)
+    white_vec = ops.matmul(prec_sqrt, ops.unsqueeze(loc, -1)).squeeze(-1)
+    return Gaussian(white_vec=white_vec, prec_sqrt=prec_sqrt, inputs=inputs)
 
 
 def random_mvn(batch_shape, dim, diag=False):
@@ -404,21 +463,25 @@ def make_plated_hmm_einsum(num_steps, num_obs_plates=1, num_hidden_plates=0):
     assert num_obs_plates >= num_hidden_plates
     t0 = num_obs_plates + 1
 
-    obs_plates = ''.join(opt_einsum.get_symbol(i) for i in range(num_obs_plates))
-    hidden_plates = ''.join(opt_einsum.get_symbol(i) for i in range(num_hidden_plates))
+    obs_plates = "".join(opt_einsum.get_symbol(i) for i in range(num_obs_plates))
+    hidden_plates = "".join(opt_einsum.get_symbol(i) for i in range(num_hidden_plates))
 
     inputs = [str(opt_einsum.get_symbol(t0))]
-    for t in range(t0, num_steps+t0):
-        inputs.append(str(opt_einsum.get_symbol(t)) + str(opt_einsum.get_symbol(t+1)) + hidden_plates)
-        inputs.append(str(opt_einsum.get_symbol(t+1)) + obs_plates)
+    for t in range(t0, num_steps + t0):
+        inputs.append(
+            str(opt_einsum.get_symbol(t))
+            + str(opt_einsum.get_symbol(t + 1))
+            + hidden_plates
+        )
+        inputs.append(str(opt_einsum.get_symbol(t + 1)) + obs_plates)
     equation = ",".join(inputs) + "->"
-    return (equation, ''.join(sorted(tuple(set(obs_plates + hidden_plates)))))
+    return (equation, "".join(sorted(tuple(set(obs_plates + hidden_plates)))))
 
 
 def make_chain_einsum(num_steps):
     inputs = [str(opt_einsum.get_symbol(0))]
     for t in range(num_steps):
-        inputs.append(str(opt_einsum.get_symbol(t)) + str(opt_einsum.get_symbol(t+1)))
+        inputs.append(str(opt_einsum.get_symbol(t)) + str(opt_einsum.get_symbol(t + 1)))
     equation = ",".join(inputs) + "->"
     return equation
 
@@ -426,7 +489,33 @@ def make_chain_einsum(num_steps):
 def make_hmm_einsum(num_steps):
     inputs = [str(opt_einsum.get_symbol(0))]
     for t in range(num_steps):
-        inputs.append(str(opt_einsum.get_symbol(t)) + str(opt_einsum.get_symbol(t+1)))
-        inputs.append(str(opt_einsum.get_symbol(t+1)))
+        inputs.append(str(opt_einsum.get_symbol(t)) + str(opt_einsum.get_symbol(t + 1)))
+        inputs.append(str(opt_einsum.get_symbol(t + 1)))
     equation = ",".join(inputs) + "->"
     return equation
+
+
+def iter_subsets(iterable, *, min_size=None, max_size=None):
+    if min_size is None:
+        min_size = 0
+    if max_size is None:
+        max_size = len(iterable)
+    for size in range(min_size, max_size + 1):
+        yield from itertools.combinations(iterable, size)
+
+
+class DesugarGetitem:
+    """
+    Helper to desugar ``.__getitem__()`` syntax.
+
+    Example::
+
+        >>> desugar_getitem[1:3, ..., None]
+        (slice(1, 3), Ellipsis, None)
+    """
+
+    def __getitem__(self, index):
+        return index
+
+
+desugar_getitem = DesugarGetitem()
