@@ -3,7 +3,9 @@
 
 import functools
 import inspect
+import pprint
 import re
+import sys
 
 import numpy as np
 
@@ -65,40 +67,97 @@ def quote(arg):
     for indent, line in out:
         if indent + len(line) >= 80:
             line += "  # noqa"
-        lines.append(' ' * indent + line)
-    return '\n'.join(lines)
+        lines.append(" " * indent + line)
+    return "\n".join(lines)
 
 
-def pretty(arg, maxlen=40):
+def pretty(arg, linewidth=75, threshold=1000):
     """
     Pretty print an expression. This is useful for debugging.
+
+    :param int linewidth: Maximum linewidth before summarization.
+    :param int threshold: Total number of array elements which trigger
+        summarization rather than full repr (default 1000). To always use the full
+        repr without summarization, pass sys.maxsize
+    :returns: A pretty string.
+    :rtype: str
     """
     out = []
-    _quote_inplace(arg, 0, out)
-    fill = u'   \u2502' * 100
+    try:
+        old = quote.printoptions.copy()
+        quote.printoptions["threshold"] = threshold
+        _quote_inplace(arg, 0, out)
+    finally:
+        quote.printoptions.update(old)
+    fill = "   \u2502" * 100
     lines = []
     for indent, line in out:
-        if len(line) > maxlen:
-            line = line[:maxlen] + "..."
+        if len(line) > linewidth:
+            line = line[: linewidth - 3] + "..."
         lines.append(fill[:indent] + line)
-    return '\n'.join(lines)
+    return "\n".join(lines)
+
+
+def _pprint_funsor(pprinter, object, stream, indent, allowance, context, level):
+    out = []
+    try:
+        old = quote.printoptions.copy()
+        quote.printoptions["threshold"] = 0  # Omit arrays from pprint.
+        _quote_inplace(object, 0, out)
+    finally:
+        quote.printoptions.update(old)
+
+    # This depends on internals of the pprint.PrettyPrinter class.
+    write = stream.write
+    for i, (extra_indent, line) in enumerate(out):
+        max_width = pprinter._width - indent - extra_indent
+        if len(line) > max_width:
+            line = line[: max_width - 3] + "..."
+        if i > 0:
+            write("\n")
+            write(" " * (indent + extra_indent))
+        write(line)
+
+
+def register_pprint(cls):
+    if hasattr(cls, "__repr__"):
+        # This depends on internals of the pprint.PrettyPrinter class.
+        pprint.PrettyPrinter._dispatch[cls.__repr__] = _pprint_funsor
 
 
 @functools.singledispatch
 def _quote_inplace(arg, indent, out):
-    line = re.sub('\n\\s*', ' ', repr(arg))
+    line = re.sub("\n\\s*", " ", repr(arg))
     out.append((indent, line))
 
 
 quote.inplace = _quote_inplace
 quote.register = _quote_inplace.register
+quote.printoptions = {"threshold": sys.maxsize}
+quote.reprtypes = ()
+
+
+def _quote_repr(arg, indent, out):
+    out.append((indent, repr(arg)))
+
+
+def _quote_register_repr(typ):
+    quote.register(typ)(_quote_repr)
+    quote.reprtypes = tuple({typ}.union(quote.reprtypes))
+
+
+quote.register_repr = _quote_register_repr
+quote.register_repr(int)
+quote.register_repr(float)
+quote.register_repr(str)
 
 
 @quote.register(tuple)
 def _(arg, indent, out):
-    if not arg:
-        out.append((indent, "()"))
+    if all(isinstance(value, quote.reprtypes) for value in arg):
+        out.append((indent, repr(arg)))
         return
+
     for value in arg[:1]:
         temp = []
         quote.inplace(value, indent + 1, temp)
@@ -106,13 +165,13 @@ def _(arg, indent, out):
         temp[0] = i - 1, "(" + line
         out.extend(temp)
         i, line = out[-1]
-        out[-1] = i, line + ','
+        out[-1] = i, line + ","
     for value in arg[1:]:
         quote.inplace(value, indent + 1, out)
         i, line = out[-1]
-        out[-1] = i, line + ','
+        out[-1] = i, line + ","
     i, line = out[-1]
-    out[-1] = i, line + ')'
+    out[-1] = i, line + ")"
 
 
 @quote.register(np.ndarray)
@@ -120,7 +179,11 @@ def _quote(arg, indent, out):
     """
     Work around NumPy ndarray not supporting reproducible repr.
     """
-    out.append((indent, "np.array({}, dtype=np.{})".format(repr(arg.tolist()), arg.dtype)))
+    if arg.size >= quote.printoptions["threshold"]:
+        data = "..." + " x ".join(str(d) for d in arg.shape) + "..."
+    else:
+        data = repr(arg.tolist())
+    out.append((indent, f"np.array({data}, dtype=np.{arg.dtype})"))
 
 
 def broadcast_shape(*shapes, **kwargs):
@@ -133,7 +196,7 @@ def broadcast_shape(*shapes, **kwargs):
     :rtype: tuple
     :raises: ValueError
     """
-    strict = kwargs.pop('strict', False)
+    strict = kwargs.pop("strict", False)
     reversed_shape = []
     for shape in shapes:
         for i, size in enumerate(reversed(shape)):
@@ -142,8 +205,11 @@ def broadcast_shape(*shapes, **kwargs):
             elif reversed_shape[i] == 1 and not strict:
                 reversed_shape[i] = size
             elif reversed_shape[i] != size and (size != 1 or strict):
-                raise ValueError('shape mismatch: objects cannot be broadcast to a single shape: {}'.format(
-                    ' vs '.join(map(str, shapes))))
+                raise ValueError(
+                    "shape mismatch: objects cannot be broadcast to a single shape: {}".format(
+                        " vs ".join(map(str, shapes))
+                    )
+                )
     return tuple(reversed(reversed_shape))
 
 
@@ -161,26 +227,34 @@ def set_backend(backend):
     :param str backend: either "numpy", "torch", or "jax".
     """
     global _FUNSOR_BACKEND, _JAX_LOADED
+    if _FUNSOR_BACKEND == backend:
+        return
 
     if backend == "numpy":
         if _JAX_LOADED:
-            raise ValueError("Cannot revert back to NumPy backend when JAX backend has been set.")
+            raise ValueError(
+                "Cannot revert back to NumPy backend when JAX backend has been set."
+            )
         else:
             _FUNSOR_BACKEND = "numpy"
     elif backend == "torch":
         _FUNSOR_BACKEND = "torch"
 
         import torch  # noqa: F401
+
         import funsor.torch  # noqa: F401
     elif backend == "jax":
         _FUNSOR_BACKEND = "jax"
         _JAX_LOADED = True
 
         import jax  # noqa: F401
+
         import funsor.jax  # noqa: F401
     else:
-        raise ValueError("backend should be either 'numpy', 'torch', or 'jax'"
-                         ", got {}".format(backend))
+        raise ValueError(
+            "backend should be either 'numpy', 'torch', or 'jax'"
+            ", got {}".format(backend)
+        )
 
 
 def get_backend():
@@ -193,23 +267,48 @@ def get_backend():
     return _FUNSOR_BACKEND
 
 
-def get_tracing_state():
-    if _FUNSOR_BACKEND == "torch":
-        import torch
+def get_default_dtype():
+    """
+    Get the current default floating point.
 
+    :return: floating point dtype.
+    :rtype: str
+    """
+    backend = get_backend()
+    if backend == "torch":
+        torch = sys.modules.get("torch")
+        return str(torch.get_default_dtype()).split(".")[1]
+    elif backend == "numpy":
+        np = sys.modules.get("numpy")
+        return np.dtype(np.float_).name
+    elif backend == "jax":
+        np = sys.modules.get("jax.numpy")
+        return np.dtype(np.float_).name
+
+
+def get_tracing_state():
+    torch = sys.modules.get("torch")
+    if torch is not None:
         return torch._C._get_tracing_state()
-    else:
-        return None
+    return None
 
 
 def is_nn_module(x):
-    if _FUNSOR_BACKEND == "torch":
-        import torch
-
+    torch = sys.modules.get("torch")
+    if torch is not None:
         return isinstance(x, torch.nn.Module)
     return False
 
 
+def as_callable(fn):
+    """
+    Converts nn.Modules ``m`` to ``m.forward``.
+    """
+    if is_nn_module(fn):
+        return fn.forward
+    return fn
+
+
 def methodof(cls, name=None):
     """
     Decorator to set the named method of the given class. Can be stacked.
@@ -220,13 +319,18 @@ def methodof(cls, name=None):
        def __call__(self, x):
            return x
     """
+
     def decorator(fn):
         name_ = name
         if name_ is None:
             fn_ = fn
             while not hasattr(fn_, "__name__"):
-                fn_ = fn_.__func__
+                if isinstance(fn_, property):
+                    fn_ = fn_.fget
+                else:
+                    fn_ = fn_.__func__
             name_ = fn_.__name__
         setattr(cls, name_, fn)
         return fn
+
     return decorator
